@@ -18,6 +18,12 @@
 #include "GL_image_display.h"
 #include "util.h"
 
+static_assert(GL_image_display_num_uniforms <= GL_image_display_max_num_uniforms,
+              "must have GL_image_display_num_uniforms <= GL_image_display_max_num_uniforms");
+static_assert(GL_image_display_num_programs <=  GL_image_display_max_num_programs,
+              "must have GL_image_display_num_programs <=  GL_image_display_max_num_programs");
+
+
 #define MAX_NUMBER_LINE_VERTICES 1024
 
 #define assert_opengl()                                 \
@@ -261,7 +267,8 @@ bool GL_image_display_init( // output
         make_uniform(aspect);
         make_uniform(center01);
         make_uniform(visible_width01);
-        make_uniform(upside_down);
+        make_uniform(flip_x);
+        make_uniform(flip_y);
         make_uniform(line_color_rgb);
 
 #undef make_uniform
@@ -416,21 +423,23 @@ bool set_aspect(GL_image_display_context_t* ctx,
     return true;
 }
 
-bool GL_image_display_update_image( GL_image_display_context_t* ctx,
-                                    int decimation_level,
+bool GL_image_display_update_image2( GL_image_display_context_t* ctx,
+                                     int decimation_level,
+                                     bool flip_x,
+                                     bool flip_y,
 
-                                    // Either this should be given
-                                    const char* image_filename,
+                                     // Either this should be given
+                                     const char* image_filename,
 
-                                    // Or these should be given
-                                    const char* image_data,
-                                    int image_width,
-                                    int image_height,
-                                    // Supported:
-                                    // - 8  for "grayscale"
-                                    // - 24 for "bgr"
-                                    int image_bpp,
-                                    int image_pitch)
+                                     // Or these should be given
+                                     const char* image_data,
+                                     int image_width,
+                                     int image_height,
+                                     // Supported:
+                                     // - 8  for "grayscale"
+                                     // - 24 for "bgr"
+                                     int image_bpp,
+                                     int image_pitch)
 {
     if(!GL_image_display_update_image__validate_input
        ( image_filename,
@@ -462,6 +471,9 @@ bool GL_image_display_update_image( GL_image_display_context_t* ctx,
         goto done;
     }
 
+    ctx->flip_x = flip_x;
+    ctx->flip_y = flip_y;
+
     if( image_filename != NULL )
     {
         FREE_IMAGE_FORMAT format = FreeImage_GetFileType(image_filename,0);
@@ -522,13 +534,14 @@ bool GL_image_display_update_image( GL_image_display_context_t* ctx,
         image_pitch  = (int)FreeImage_GetPitch(fib);
         image_data   = (char*)FreeImage_GetBits(fib);
 
-        // FreeImage_Load() loads images upside down
-        ctx->upside_down = true;
+        // FreeImage_Load() loads images upside down, so I tell OpenGL to do the
+        // opposite thing, in terms of flipping stuff upside down. The REST of
+        // the logic regarding flip_y stays the same: I do NOT flip ctx->flip_y
+        set_uniform_1i(ctx, flip_y, !ctx->flip_y);
     }
     else
-        ctx->upside_down = false;
-
-    set_uniform_1i(ctx, upside_down, ctx->upside_down);
+        set_uniform_1i(ctx, flip_y, ctx->flip_y);
+    set_uniform_1i(ctx, flip_x, ctx->flip_x);
 
     if(!ctx->did_init_texture)
     {
@@ -692,6 +705,25 @@ bool GL_image_display_update_image( GL_image_display_context_t* ctx,
     return result;
 }
 
+bool GL_image_display_update_image( GL_image_display_context_t* ctx,
+                                    int decimation_level,
+                                    const char* image_filename,
+                                    const char* image_data,
+                                    int image_width,
+                                    int image_height,
+                                    int image_bpp,
+                                    int image_pitch)
+{
+    return GL_image_display_update_image2( ctx,
+                                           decimation_level,
+                                           false,false,
+                                           image_filename,
+                                           image_data,
+                                           image_width,
+                                           image_height,
+                                           image_bpp,
+                                           image_pitch);
+}
 
 void GL_image_display_deinit( GL_image_display_context_t* ctx )
 {
@@ -733,7 +765,7 @@ bool GL_image_display_set_panzoom(GL_image_display_context_t* ctx,
     CONFIRM_SET(did_init_texture);
 
 #define TRY_EXISTING_OR_SET(what)                                                \
-    /* check for isfinite() AND big values because -ffast-mathj breaks isfinite()*/ \
+    /* check for isfinite() AND big values because -ffast-math breaks isfinite()*/ \
     if( !isfinite(what) || what >= 1e20 || what <= -1e20 )                       \
     {                                                                            \
         /* Invalid input. I keep existing value IF there is an existing value */ \
@@ -915,12 +947,11 @@ bool GL_image_display_map_pixel_viewport_from_image(GL_image_display_context_t*
     double vertex_x = (x+0.5) / ((double)(1 << ctx->decimation_level)*(double)ctx->image_width);
     double vertex_y = (y+0.5) / ((double)(1 << ctx->decimation_level)*(double)ctx->image_height);
 
-    // GL does things upside down. It looks like this should be unconditional,
-    // independent of ctx->upside_down. The shader has an if(). I'm not
-    // completely sure why this is so, but tests say that it is, and I'm
-    // confident that if I think hard enough I will convince myself that it is
-    // right
-    vertex_y = 1.0 - vertex_y;
+    // GL does things upside down so the logic on vertex_y has the opposite polarity
+    if(ctx->flip_x)
+        vertex_x = 1.0 - vertex_x;
+    if(!ctx->flip_y)
+        vertex_y = 1.0 - vertex_y;
 
     // gl_Position. In [-1,1]
     double glpos_x =
@@ -967,12 +998,11 @@ bool GL_image_display_map_pixel_image_from_viewport(GL_image_display_context_t*
     double vertex_y =
         glpos_y / (2. * ctx->aspect_y) * ctx->visible_width01 + ctx->center01_y;
 
-    // GL does things upside down. It looks like this should be unconditional,
-    // independent of ctx->upside_down. The shader has an if(). I'm not
-    // completely sure why this is so, but tests say that it is, and I'm
-    // confident that if I think hard enough I will convince myself that it is
-    // right
-    vertex_y = 1.0 - vertex_y;
+    // GL does things upside down so the logic on vertex_y has the opposite polarity
+    if(ctx->flip_x)
+        vertex_x = 1.0 - vertex_x;
+    if(!ctx->flip_y)
+        vertex_y = 1.0 - vertex_y;
 
     *xout = vertex_x*(double)(1 << ctx->decimation_level)*(double)ctx->image_width  - 0.5;
     *yout = vertex_y*(double)(1 << ctx->decimation_level)*(double)ctx->image_height - 0.5;
